// 8.04.2019, Arduino IDE v1.8.9, LZ2WSG, KN34PC
// Si5351 VFO CLK0, OLED 0.91" 128x32 display, S-meter
//---------------------------------------------------------------------------------------------------------
#include "src/Rotary.h"              // http://www.buxtronix.net/2011/10/rotary-encoders-done-properly.html
#include "src/si5351.h"              // https://github.com/etherkit/Si5351Arduino, v2.1.0
#include "src/SSD1306AsciiAvrI2c.h"  // https://github.com/greiman/SSD1306Ascii

#define DEFAULT_RX    7100000
#define BAND_L        7000000
#define BAND_H        7200000
#define IF_FREQ       496500
#define PIN_SM        A0            // S-meter voltage input
#define SW_STEP       4             // encoder step switch
#define PIN_RX_TX     5             // RX/TX input
#define ENC_INC_1     50            // encoder increment 1
#define ENC_INC_2     500           // encoder increment 2
#define OLED_I2C_ADDR 0x3C

#define PIN_KEY_LINE       6  // digital pin for the key line (mirrors PIN_TONE_OUT)
#define PIN_PTT_LINE       7  // digital pin for the PTT line
#define PIN_TONE_OUT       8  // digital pin with keyed audio tone on it / use 8 ohm speaker with 220 ohm series resistor to earth
#define PIN_PADDLE_R      11  // digital pin for paddle left (dot)
#define PIN_PADDLE_L      12  // digital pin for paddle right (dash)
#define PIN_KEYER_SPEED   A3  // analogue pin for speed potentiometer  
#define CW_TONE_HZ       700  // CW tone frequency (Hz)  
#define CW_DASH_LEN       4  // length of dash (in dots)
#define BREAK_IN_DELAY   800  // break-in hang time (mS)


Rotary r = Rotary(2, 3);
Si5351 si5351(0x60);                 // Si5351 I2C address
SSD1306AsciiAvrI2c oled;

uint32_t rx;
uint32_t enc_inc;
int8_t enc_dir = 0;                  // -1 DIR_CCW, 0 DIR_NONE, 1 DIR_CCW
bool ch_flag = true;
uint16_t val_sm = 0;
uint16_t val_sm_old = 1;
bool rx_tx;
bool rx_tx_old;
bool tx_new = false;

enum trx_state_e {
  E_STATE_RX, 
  E_STATE_TX
};  // state of the txcvr 

trx_state_e curr_state;

enum key_state_e {
  E_KEY_UP, 
  E_KEY_DOWN
};  // state of the key line 
key_state_e key_state;

int char_sent_ms, curr_ms;
int dot_length_ms = 60;   // keyer base speed (60 equates to 10 w.p.m.)
int last_ident_ms;        // CW ident timer
int ident_secs_count = 0; // seconds counter
                         
bool space_inserted;
bool paddle_squeezed;
unsigned int ch_counter;



bool get_button(byte btn)
{
// Read the digital pin 'btn', with debouncing  
// pins are pulled up (ie. default is HIGH)
// returns TRUE if pin is high, FALSE otherwise 
  if (!digitalRead(btn)) 
  {
    delay(5);  // was 20mS, needs to be long enough to ensure debouncing
    if (!digitalRead(btn))
    {
      // while (!digitalRead(btn));  // loop here until button released
      return 1;                   // now, return 'true' 
    }
  }
  return 0;
}



int read_analogue_pin(byte p)
{
// Take an averaged reading of analogue pin 'p'  
  int i, val=0, nbr_reads=2; 
  for (i=0; i<nbr_reads; i++)
  {
    val += analogRead(p);
    delay(5); 
  }
  return val/nbr_reads; 
}

int read_keyer_speed()
{ 
  int n = read_analogue_pin((byte)PIN_KEYER_SPEED);
  dot_length_ms = 60 + (n-183)/5;   // scale to wpm (10 wpm == 60mS dot length)
                                     // '511' should be mid point of returned range
                                     // change '5' to widen/narrow speed range...
                                     // smaller number -> greater range  
  return n;
}



void set_key_state(key_state_e k)
{
// push the morse key down, or let it spring back up
// changes global 'key_state' {E_KEY_DOWN, E_KEY_UP}

  switch (k)
  {
      case E_KEY_DOWN:
      {
        // do whatever you need to key the transmitter
        digitalWrite(PIN_KEY_LINE, LOW);
        digitalWrite(13, 1);  // for now, turn the Nano's LED on
        key_state = E_KEY_DOWN;
      }
      break;

      case E_KEY_UP:
      {
        // do whatever you need to un-key the transmitter 
        digitalWrite(PIN_KEY_LINE, HIGH);
        digitalWrite(13, 0);  // for now, turn the Nano's LED off
        key_state = E_KEY_UP;
        char_sent_ms = millis();
        space_inserted = false;
      }
      break;
  }    
}

void activate_state(trx_state_e s)
{
// if necessary, activate the receiver or the transmitter 
// changes global 'curr_state' {E_STATE_RX, E_STATE_TX}
  switch (s)
  {
      case E_STATE_RX:
      {
        if(curr_state == E_STATE_RX)
        {
          // already in receive state, nothing to do!
        }
        else
        { 
          // turn transmitter off (drop PTT line)
          digitalWrite(PIN_PTT_LINE, 0); 
          // un-mute receiver
          curr_state = E_STATE_RX;
        }
      }
      break;

      case E_STATE_TX:
      {
        if(curr_state == E_STATE_TX)
        {
          // already in transmit state, nothing to do!
        }
        else
        {
          // turn transmitter on (raise PTT line) 
          digitalWrite(PIN_PTT_LINE, 1); 
          curr_state = E_STATE_TX;
        }
      }
      break;
  }    
}

void send_dot()
{
  delay(dot_length_ms);  // wait for one dot period (space)

  // send a dot and the following space
  tone(PIN_TONE_OUT, CW_TONE_HZ);

  si5351.set_freq((rx-800) * SI5351_FREQ_MULT, SI5351_CLK2);
  si5351.output_enable(SI5351_CLK2, 1);
  
  set_key_state(E_KEY_DOWN);
  delay(dot_length_ms);  // key down for one dot period
  noTone(PIN_TONE_OUT);

  si5351.set_freq((rx-800) * SI5351_FREQ_MULT, SI5351_CLK2);
  si5351.output_enable(SI5351_CLK2, 0);

  
  set_key_state(E_KEY_UP); 
  ch_counter++;
}

void send_dash()
{
  delay(dot_length_ms);  // wait for one dot period (space)
  // send a dash and the following space
  tone(PIN_TONE_OUT, CW_TONE_HZ);

  si5351.set_freq((rx-800) * SI5351_FREQ_MULT, SI5351_CLK2);
  si5351.output_enable(SI5351_CLK2, 1);

  
  set_key_state(E_KEY_DOWN); 
  delay(dot_length_ms * CW_DASH_LEN);  // key down for CW_DASH_LEN dot periods
  noTone(PIN_TONE_OUT);


 si5351.set_freq((rx-800) * SI5351_FREQ_MULT, SI5351_CLK2);
 si5351.output_enable(SI5351_CLK2, 0);

  
  set_key_state(E_KEY_UP); 
  ch_counter++;
}

void send_letter_space()
{
  delay(dot_length_ms * 4);  // wait for 3 dot periods
}

void send_word_space()
{
  delay(dot_length_ms * 7);  // wait for 6 dot periods
}

void send_morse_char(char c)
{
  // 'c' is a '.' or '-' char, so send it 
  if(c == '.') send_dot();
  else if (c == '-') send_dash();
  // ignore anything else, including 0s
}







//---------------------------------------------------------------------------------------------------------
void setup() {
  pinMode(2, INPUT_PULLUP);
  pinMode(3, INPUT_PULLUP);
  pinMode(4, INPUT_PULLUP);
  pinMode(5, INPUT_PULLUP);
  pinMode(13, OUTPUT);
  pinMode(6, OUTPUT);
  pinMode(7, OUTPUT); 
  pinMode(PIN_PADDLE_L,  INPUT_PULLUP);
  pinMode(PIN_PADDLE_R,  INPUT_PULLUP);
  digitalWrite(PIN_KEY_LINE, HIGH);
  digitalWrite(PIN_PTT_LINE, 0);  // send the PTT line low

  analogReference(INTERNAL);      // ADC reference voltage 5,0V or 1,1V (S-meter max IN voltage)

  attachInterrupt(0, rotary_encoder, CHANGE);
  attachInterrupt(1, rotary_encoder, CHANGE);

si5351.init(SI5351_CRYSTAL_LOAD_8PF, 0, 61000);    // crystal 25.000 MHz, correction -610 Hz

  si5351.output_enable(SI5351_CLK0, 1);
  si5351.output_enable(SI5351_CLK1, 1);
  si5351.output_enable(SI5351_CLK2, 1);

  si5351.drive_strength(SI5351_CLK0, SI5351_DRIVE_8MA);
  si5351.drive_strength(SI5351_CLK1, SI5351_DRIVE_8MA);
  si5351.drive_strength(SI5351_CLK2, SI5351_DRIVE_8MA);

  si5351.set_freq(IF_FREQ * SI5351_FREQ_MULT, SI5351_CLK1);

  oled.begin(&Adafruit128x32, OLED_I2C_ADDR);

  rx = DEFAULT_RX;

  rx_tx = digitalRead(PIN_RX_TX);
  rx_tx_old = !rx_tx;
  show_state();

  enc_inc = ENC_INC_2;
  show_inc();



  curr_state = E_STATE_RX;
  key_state = E_KEY_UP;
  char_sent_ms = millis();
  space_inserted = false;
  paddle_squeezed = false;
  ch_counter = 0;

  last_ident_ms = millis(); 



}
//---------------------------------------------------------------------------------------------------------
void loop() {
  if (ch_flag) {
    ch_flag = false;
    rx += enc_dir * enc_inc;
    enc_dir = 0;
    rx = constrain(rx, BAND_L, BAND_H);

    si5351.set_freq((rx + IF_FREQ) * SI5351_FREQ_MULT, SI5351_CLK0);

    show_freq(rx);
    tx_new = true;
  }

  if (digitalRead(SW_STEP) == LOW) { // increment events
    if (button_state(SW_STEP) > 0) {
      set_inc();
      show_inc();
      do {
      }
      while (digitalRead(SW_STEP) == LOW);
    }
  }

  rx_tx = digitalRead(PIN_RX_TX);
  if (rx_tx != rx_tx_old) {
    if (tx_new) {
       si5351.set_freq((rx-800) * SI5351_FREQ_MULT, SI5351_CLK2);
      tx_new = false;
    }
    si5351.output_enable(SI5351_CLK2, !rx_tx);
    show_state();
    rx_tx_old = rx_tx;
  }

  val_sm = analogRead(PIN_SM);
  //val_sm = log(val_sm);
 // val_sm = 180*val_sm;
  val_sm = val_sm/1.0;
  //if ((abs(val_sm - val_sm_old)) > 200) {  // if needed draw S-meter
    show_sm();
    val_sm_old = val_sm;
  //}

//-- keyer code begins --------------------------------------------------
  // read the speed control 
  read_keyer_speed();

  // see if the paddle has been pressed
  bool l_paddle = get_button(PIN_PADDLE_L);
  bool r_paddle = get_button(PIN_PADDLE_R);

  if(l_paddle or r_paddle) activate_state(E_STATE_TX);
  if(l_paddle) send_dot();
  if(r_paddle) send_dash();
  if(l_paddle and r_paddle) // paddle_squeezed = true;
  {
     send_dot();
     send_dash();
  }  

  curr_ms = millis();
  // if paddle has been idle for BREAK_IN_DELAY drop out of transmit 
  if((curr_state == E_STATE_TX) and (curr_ms - char_sent_ms) > BREAK_IN_DELAY)
  {
    // drop back to receive to implement break-in
    activate_state(E_STATE_RX); 
  }
  //-- keyer code ends --------------------------------------------------

}
//---------------------------------------------------------------------------------------------------------
void rotary_encoder() {              // rotary encoder events
  uint8_t result = r.process();

  if (result) {
    ch_flag = true;
    if (result == DIR_CW)
      enc_dir = 1;
    else
      enc_dir = -1;
  }
}
//---------------------------------------------------------------------------------------------------------
void set_inc() {
  if (enc_inc == ENC_INC_1)
    enc_inc = ENC_INC_2;             // 500 Hz
  else
    enc_inc = ENC_INC_1;             // 50 Hz
}
//---------------------------------------------------------------------------------------------------------
void show_inc() {
  oled.setCursor(109, 3);
  oled.setFont(labels);
  if (enc_inc == ENC_INC_1)
    oled.print("6");                 // 50 Hz
  else
    oled.print("9");                 // 500 Hz
}
//---------------------------------------------------------------------------------------------------------
void show_freq(uint32_t sf_rx) {
  uint16_t fr;

  oled.setFont(lcdnums14x24);
  oled.setCursor(1, 0);
  fr = sf_rx / 1000000;
  if (fr < 10)
    oled.print(':');                 // ':' is changed to ' ' in lcdnums14x24.h
  oled.print(fr);
  oled.print('.');
  fr = (sf_rx % 1000000) / 1000;
  if (fr < 100)
    oled.print('0');
  if (fr < 10)
    oled.print('0');
  oled.print(fr);
  oled.print('.');
  fr = (sf_rx % 1000) / 10;
  if (fr < 10)
    oled.print('0');
  oled.print(fr);
}
//---------------------------------------------------------------------------------------------------------
void show_state() {
  oled.setCursor(2, 3);
  oled.setFont(labels);
  if (rx_tx)
    {oled.print("1");                 // "1" is "RX" in labels.h
     digitalWrite(13, LOW);}
  else
    {oled.print("2");                 // "2" is "TX" in labels.h
     digitalWrite(13, HIGH);}
}
//---------------------------------------------------------------------------------------------------------
void show_sm() {
  uint8_t a = 0;
  uint8_t m = 0;

  a = (val_sm + 3) / 113;            // 1024 / 9 characters for S = 1,3,5,7,8,9,+10,+20,+30
  oled.setFont(pixels);
  oled.setCursor(24, 3);
  for (m = 0; m < a; m++)
    if (m < 6)
      oled.print('7');               // '5' - hollow rectangle, 6px
    else
      oled.print('8');               // '6' - filled rectangle, 6px
  for (m = a; m < 9; m++)
    oled.print('.');                 // '.' 1px
}
//---------------------------------------------------------------------------------------------------------
uint8_t button_state(uint8_t pin_bt) {  // push button state: 0(very fast), 1(fast), 2(slow)
  uint32_t t_start;
  uint32_t t_bt = 0;
  uint8_t bt_state;

  t_start = millis();
  while (!digitalRead(pin_bt)) {
    t_bt = millis() - t_start;
    if (t_bt > 1000)
      break;
  }
  if (t_bt < 20)                     // 'nothing' for debounce
    bt_state = 0;
  else if (t_bt < 400)               // fast
    bt_state = 1;
  else                               // slow
    bt_state = 2;
  return bt_state;
}






